@@ -25,3 +25,13 @@ class SoftTimeLimitExceeded(Exception):
 
 class WorkerLostError(Exception):
     """The worker processing a job has exited prematurely."""
+
+
+class RestartFreqExceeded(Exception):
+    """Restarts too fast."""
+
+
+class CoroStop(Exception):
+    """Coroutine exit, as opposed to StopIteration which may
+    mean it should be restarted."""
+    pass
