@@ -53,10 +53,13 @@ import warnings
 
 from . import Event, Process, cpu_count
 from . import util
+from .common import restart_state
 from .einfo import ExceptionInfo
 from .exceptions import (
+    CoroStop,
     SoftTimeLimitExceeded,
     TimeLimitExceeded,
+    RestartFreqExceeded,
     TimeoutError,
     WorkerLostError,
 )
@@ -87,6 +90,7 @@ SIG_SOFT_TIMEOUT = getattr(signal, "SIGUSR1", None)
 #
 
 LOST_WORKER_TIMEOUT = 10.0
+EX_OK = getattr(os, "EX_OK", 0)
 
 job_counter = itertools.count()
 
@@ -108,11 +112,16 @@ def safe_apply_callback(fun, *args):
     if fun:
         try:
             fun(*args)
-        except BaseException, exc:
+        except Exception, exc:
             error("Pool callback raised exception: %r", exc,
                   exc_info=True)
 
 
+def stop_if_not_current(thread, timeout=None):
+    if thread is not threading.current_thread():
+        thread.stop(timeout)
+
+
 class LaxBoundedSemaphore(threading._Semaphore):
     """Semaphore that checks that # release is <= # acquires,
     but ignores if # releases >= value."""
@@ -121,14 +130,33 @@ class LaxBoundedSemaphore(threading._Semaphore):
         _Semaphore.__init__(self, value, verbose)
         self._initial_value = value
 
+    def grow(self):
+        cond = self._Semaphore__cond
+        with cond:
+            self._initial_value += 1
+            self._Semaphore__value += 1
+            cond.notify()
+
+    def shrink(self, callback):
+        self._initial_value -= 1
+        self.acquire()
+
     if sys.version_info[0] == 3:
 
         def release(self):
-            if self._value < self._initial_value:
-                _Semaphore.release(self)
-            if __debug__:
-                self._note("%s.release: success, value=%s (unchanged)" % (
-                    self, self._value))
+            cond = self._Semaphore__cond
+            with cond:
+                if self._value < self._initial_value:
+                    self._value += 1
+                    cond.notify_all()
+                    if __debug__:
+                        self._note("%s.release: success, value=%s",
+                            self, self._value)
+                else:
+                    if __debug__:
+                        self._note(
+                            "%s.release: success, value=%s (unchanged)" % (
+                                self, self._value))
 
         def clear(self):
             while self._value < self._initial_value:
@@ -136,11 +164,19 @@ class LaxBoundedSemaphore(threading._Semaphore):
     else:
 
         def release(self):  # noqa
-            if self._Semaphore__value < self._initial_value:
-                _Semaphore.release(self)
-            if __debug__:
-                self._note("%s.release: success, value=%s (unchanged)" % (
-                    self, self._Semaphore__value))
+            cond = self._Semaphore__cond
+            with cond:
+                if self._Semaphore__value < self._initial_value:
+                    self._Semaphore__value += 1
+                    cond.notifyAll()
+                    if __debug__:
+                        self._note("%s.release: success, value=%s",
+                            self, self._Semaphore__value)
+                else:
+                    if __debug__:
+                        self._note(
+                            "%s.release: success, value=%s (unchanged)" % (
+                                self, self._Semaphore__value))
 
         def clear(self):  # noqa
             while self._Semaphore__value < self._initial_value:
@@ -270,16 +306,35 @@ class PoolThread(threading.Thread):
     def __init__(self, *args, **kwargs):
         threading.Thread.__init__(self)
         self._state = RUN
+        self._was_started = False
         self.daemon = True
 
     def run(self):
         try:
             return self.body()
+        except RestartFreqExceeded, exc:
+            error("Thread %r crashed: %r", type(self).__name__, exc,
+                  exc_info=True)
+            os.kill(os.getpid(), signal.SIGTERM)
+            sys.exit()
         except Exception, exc:
             error("Thread %r crashed: %r", type(self).__name__, exc,
                   exc_info=True)
             os._exit(1)
 
+    def start(self, *args, **kwargs):
+        self._was_started = True
+        super(PoolThread, self).start(*args, **kwargs)
+
+    def on_stop_not_started(self):
+        pass
+
+    def stop(self, timeout=None):
+        if self._was_started:
+            self.join(timeout)
+            return
+        self.on_stop_not_started()
+
     def terminate(self):
         self._state = TERMINATE
 
@@ -296,11 +351,31 @@ class Supervisor(PoolThread):
     def body(self):
         debug('worker handler starting')
 
-        # Keep maintaing workers until the cache gets drained, unless the pool
-        # is termianted
-        while self._state == RUN and self.pool._state == RUN:
-            self.pool._maintain_pool()
-            time.sleep(0.8)
+        time.sleep(0.8)
+
+        pool = self.pool
+
+        try:
+            # do a burst at startup to verify that we can start
+            # our pool processes, and in that time we lower
+            # the max restart frequency.
+            prev_state = pool.restart_state
+            pool.restart_state = restart_state(10 * pool._processes, 1)
+            for _ in xrange(10):
+                if self._state == RUN and pool._state == RUN:
+                    pool._maintain_pool()
+                    time.sleep(0.1)
+
+            # Keep maintaing workers until the cache gets drained, unless
+            # the pool is termianted
+            pool.restart_state = prev_state
+            while self._state == RUN and pool._state == RUN:
+                pool._maintain_pool()
+                time.sleep(0.8)
+        except RestartFreqExceeded:
+            pool.close()
+            pool.join()
+            raise
         debug('worker handler exiting')
 
 
@@ -315,9 +390,7 @@ class TaskHandler(PoolThread):
 
     def body(self):
         taskqueue = self.taskqueue
-        outqueue = self.outqueue
         put = self.put
-        pool = self.pool
 
         for taskseq, set_length in iter(taskqueue.get, None):
             i = -1
@@ -339,6 +412,13 @@ class TaskHandler(PoolThread):
         else:
             debug('task handler got sentinel')
 
+        self.tell_others()
+
+    def tell_others(self):
+        outqueue = self.outqueue
+        put = self.put
+        pool = self.pool
+
         try:
             # tell result handler to finish when cache is empty
             debug('task handler sending sentinel to result handler')
@@ -353,6 +433,9 @@ class TaskHandler(PoolThread):
 
         debug('task handler exiting')
 
+    def on_stop_not_started(self):
+        self.tell_others()
+
 
 class TimeoutHandler(PoolThread):
 
@@ -450,22 +533,27 @@ class TimeoutHandler(PoolThread):
 class ResultHandler(PoolThread):
 
     def __init__(self, outqueue, get, cache, poll,
-            join_exited_workers, putlock):
+            join_exited_workers, putlock, restart_state):
         self.outqueue = outqueue
         self.get = get
         self.cache = cache
         self.poll = poll
         self.join_exited_workers = join_exited_workers
         self.putlock = putlock
+        self.restart_state = restart_state
+        self._it = None
+        self._shutdown_complete = False
         super(ResultHandler, self).__init__()
 
-    def body(self):
-        get = self.get
-        outqueue = self.outqueue
+    def on_stop_not_started(self):
+        # used when pool started without result handler thread.
+        self.finish_at_shutdown()
+
+    def _process_result(self, timeout=1.0):
         cache = self.cache
         poll = self.poll
-        join_exited_workers = self.join_exited_workers
         putlock = self.putlock
+        restart_state = self.restart_state
 
         def on_ack(job, i, time_accepted, pid):
             try:
@@ -475,6 +563,7 @@ class ResultHandler(PoolThread):
                 pass
 
         def on_ready(job, i, obj):
+            restart_state.R = 0
             try:
                 item = cache[job]
             except KeyError:
@@ -496,25 +585,89 @@ class ResultHandler(PoolThread):
             except KeyError:
                 debug("Unknown job state: %s (args=%s)", state, args)
 
-        debug('result handler starting')
         while 1:
             try:
-                ready, task = poll(1.0)
+                ready, task = poll(timeout)
             except (IOError, EOFError), exc:
                 debug('result handler got %r -- exiting', exc)
-                return
+                raise CoroStop()
 
             if self._state:
                 assert self._state == TERMINATE
                 debug('result handler found thread._state=TERMINATE')
-                break
+                raise CoroStop()
 
             if ready:
                 if task is None:
                     debug('result handler got sentinel')
-                    break
-
+                    raise CoroStop()
                 on_state_change(task)
+                if timeout != 0:  # blocking
+                    break
+            else:
+                break
+
+        yield
+
+    def handle_event(self, fileno, event):
+        if self._state == RUN:
+            if self._it is None:
+                self._it = self._process_result(0)  # non-blocking
+            try:
+                self._it.next()
+            except StopIteration:
+                self._it = None
+
+    def body(self):
+        debug('result handler starting')
+        try:
+            while 1:
+                for _ in self._process_result(1.0):  # blocking
+                    pass
+        except CoroStop:
+            return
+        finally:
+            self.finish_at_shutdown()
+
+    def finish_at_shutdown(self):
+        self._shutdown_complete = True
+        get = self.get
+        outqueue = self.outqueue
+        cache = self.cache
+        poll = self.poll
+        join_exited_workers = self.join_exited_workers
+        putlock = self.putlock
+        restart_state = self.restart_state
+
+        def on_ack(job, i, time_accepted, pid):
+            try:
+                cache[job]._ack(i, time_accepted, pid)
+            except (KeyError, AttributeError):
+                # Object gone or doesn't support _ack (e.g. IMAPIterator).
+                pass
+
+        def on_ready(job, i, obj):
+            restart_state.R = 0
+            try:
+                item = cache[job]
+            except KeyError:
+                return
+            if not item.ready():
+                if putlock is not None:
+                    putlock.release()
+            try:
+                item._set(i, obj)
+            except KeyError:
+                pass
+
+        state_handlers = {ACK: on_ack, READY: on_ready}
+
+        def on_state_change(task):
+            state, args = task
+            try:
+                state_handlers[state](*args)
+            except KeyError:
+                debug("Unknown job state: %s (args=%s)", state, args)
 
         time_terminate = None
         while cache and self._state != TERMINATE:
@@ -574,7 +727,14 @@ class Pool(object):
 
     def __init__(self, processes=None, initializer=None, initargs=(),
             maxtasksperchild=None, timeout=None, soft_timeout=None,
-            lost_worker_timeout=LOST_WORKER_TIMEOUT):
+            lost_worker_timeout=LOST_WORKER_TIMEOUT,
+            max_restarts=None, max_restart_freq=1,
+            on_process_up=None,
+            on_process_down=None,
+            with_task_thread=True,
+            with_result_thread=True,
+            with_supervisor_thread=True,
+            semaphore=None):
         self._setup_queues()
         self._taskqueue = Queue.Queue()
         self._cache = {}
@@ -585,6 +745,14 @@ class Pool(object):
         self._initializer = initializer
         self._initargs = initargs
         self.lost_worker_timeout = lost_worker_timeout or LOST_WORKER_TIMEOUT
+        self.max_restarts = max_restarts or round(processes * 100)
+        self.restart_state = restart_state(max_restarts, max_restart_freq or 1)
+        self.on_process_up = on_process_up
+        self.on_process_down = on_process_down
+        self.with_task_thread = with_task_thread
+        self.with_result_thread = with_result_thread
+        self.with_supervisor_thread = with_supervisor_thread
+        self.readers = {}
 
         if soft_timeout and SIG_SOFT_TIMEOUT is None:
             warnings.warn(UserWarning("Soft timeouts are not supported: "
@@ -603,18 +771,24 @@ class Pool(object):
 
         self._pool = []
         self._poolctrl = {}
-        self._putlock = LaxBoundedSemaphore(self._processes)
+        self._putlock = semaphore or LaxBoundedSemaphore(self._processes)
         for i in range(processes):
             self._create_worker_process()
 
         self._worker_handler = self.Supervisor(self)
-        self._worker_handler.start()
+        if with_supervisor_thread:
+            self._worker_handler.start()
+        else:
+            self.readers.update(
+                dict((w._popen.sentinel, self.maintain_pool)
+                        for w in self._pool))
 
         self._task_handler = self.TaskHandler(self._taskqueue,
                                               self._quick_put,
                                               self._outqueue,
                                               self._pool)
-        self._task_handler.start()
+        if with_task_thread:
+            self._task_handler.start()
 
         # Thread killing timedout jobs.
         self._timeout_handler = None
@@ -627,8 +801,13 @@ class Pool(object):
                                         self._quick_get, self._cache,
                                         self._poll_result,
                                         self._join_exited_workers,
-                                        self._putlock)
-        self._result_handler.start()
+                                        self._putlock,
+                                        self.restart_state)
+        if with_result_thread:
+            self._result_handler.start()
+        else:
+            self.readers[self._outqueue._reader] = \
+                    self._result_handler.handle_event
 
         self._terminate = Finalize(
             self, self._terminate_pool,
@@ -652,6 +831,8 @@ class Pool(object):
         w.name = w.name.replace('Process', 'PoolWorker')
         w.daemon = True
         w.start()
+        if self.on_process_up:
+            self.on_process_up(w)
         self._poolctrl[w.pid] = sentinel
         return w
 
@@ -668,9 +849,12 @@ class Pool(object):
         for job in [job for job in self._cache.values()
                 if not job.ready() and job._worker_lost]:
             now = now or time.time()
-            if now - job._worker_lost > job._lost_worker_timeout:
+            lost_time, lost_ret = job._worker_lost
+            if now - lost_time > job._lost_worker_timeout:
                 try:
-                    raise WorkerLostError("Worker exited prematurely.")
+                    raise WorkerLostError(
+                        "Worker exited prematurely (exitcode: %r)." % (
+                            lost_ret, ))
                 except WorkerLostError:
                     exc_info = ExceptionInfo()
                     job._set(None, (False, exc_info))
@@ -680,7 +864,7 @@ class Pool(object):
         if shutdown and not len(self._pool):
             raise WorkersJoined()
 
-        cleaned = []
+        cleaned, exitcodes = [], {}
         for i in reversed(range(len(self._pool))):
             worker = self._pool[i]
             if worker.exitcode is not None:
@@ -689,26 +873,28 @@ class Pool(object):
                 worker.join()
                 debug('Supervisor: worked %d joined', i)
                 cleaned.append(worker.pid)
+                exitcodes[worker.pid] = worker.exitcode
                 del self._pool[i]
                 del self._poolctrl[worker.pid]
+                if self.on_process_down:
+                    self.on_process_down(worker)
         if cleaned:
             for job in self._cache.values():
                 for worker_pid in job.worker_pids():
                     if worker_pid in cleaned and not job.ready():
-                        job._worker_lost = time.time()
+                        job._worker_lost = (time.time(), exitcodes[worker_pid])
                         continue
             if self._putlock is not None:
                 for worker in cleaned:
                     self._putlock.release()
-            return True
+            return exitcodes.values()
         return False
 
     def shrink(self, n=1):
         for i, worker in enumerate(self._iterinactive()):
             self._processes -= 1
             if self._putlock:
-                self._putlock._initial_value -= 1
-                self._putlock.acquire()
+                self._putlock.shrink()
             worker.terminate()
             if i == n - 1:
                 return
@@ -719,11 +905,7 @@ class Pool(object):
             #assert len(self._pool) == self._processes
             self._processes += 1
             if self._putlock:
-                cond = self._putlock._Semaphore__cond
-                with cond:
-                    self._putlock._initial_value += 1
-                    self._putlock._Semaphore__value += 1
-                    cond.notify()
+                self.putlock.grow()
 
     def _iterinactive(self):
         for worker in self._pool:
@@ -737,21 +919,37 @@ class Pool(object):
                 return True
         return False
 
-    def _repopulate_pool(self):
+    def _repopulate_pool(self, exitcodes):
         """Bring the number of pool processes up to the specified number,
         for use after reaping workers which have exited.
         """
         for i in range(self._processes - len(self._pool)):
             if self._state != RUN:
                 return
+            try:
+                if exitcodes[i] != EX_OK:
+                    self.restart_state.step()
+            except IndexError:
+                self.restart_state.step()
             self._create_worker_process()
             debug('added worker')
 
+    def did_start_ok(self):
+        return not self._join_exited_workers()
+
     def _maintain_pool(self):
         """"Clean up any exited workers and start replacements for them.
         """
-        self._join_exited_workers()
-        self._repopulate_pool()
+        self._repopulate_pool(self._join_exited_workers())
+
+    def maintain_pool(self, *args, **kwargs):
+        if self._worker_handler._state == RUN and self._state == RUN:
+            try:
+                self._maintain_pool()
+            except RestartFreqExceeded:
+                self.close()
+                self.join()
+                raise
 
     def _setup_queues(self):
         from billiard.queues import SimpleQueue
@@ -780,8 +978,8 @@ class Pool(object):
         '''
         Equivalent of `func(*args, **kwargs)`.
         '''
-        assert self._state == RUN
-        return self.apply_async(func, args, kwds).get()
+        if self._state == RUN:
+            return self.apply_async(func, args, kwds).get()
 
     def starmap(self, func, iterable, chunksize=None):
         '''
@@ -789,31 +987,33 @@ class Pool(object):
         be iterables as well and will be unpacked as arguments. Hence
         `func` and (a, b) becomes func(a, b).
         '''
-        assert self._state == RUN
-        return self._map_async(func, iterable, starmapstar, chunksize).get()
+        if self._state == RUN:
+            return self._map_async(func, iterable,
+                                   starmapstar, chunksize).get()
 
     def starmap_async(self, func, iterable, chunksize=None, callback=None,
             error_callback=None):
         '''
         Asynchronous version of `starmap()` method.
         '''
-        assert self._state == RUN
-        return self._map_async(func, iterable, starmapstar, chunksize,
-                               callback, error_callback)
+        if self._state == RUN:
+            return self._map_async(func, iterable, starmapstar, chunksize,
+                                   callback, error_callback)
 
     def map(self, func, iterable, chunksize=None):
         '''
         Apply `func` to each element in `iterable`, collecting the results
         in a list that is returned.
         '''
-        assert self._state == RUN
-        return self.map_async(func, iterable, chunksize).get()
+        if self._state == RUN:
+            return self.map_async(func, iterable, chunksize).get()
 
     def imap(self, func, iterable, chunksize=1, lost_worker_timeout=None):
         '''
         Equivalent of `map()` -- can be MUCH slower than `Pool.map()`.
         '''
-        assert self._state == RUN
+        if self._state != RUN:
+            return
         lost_worker_timeout = lost_worker_timeout or self.lost_worker_timeout
         if chunksize == 1:
             result = IMapIterator(self._cache,
@@ -835,7 +1035,8 @@ class Pool(object):
         '''
         Like `imap()` method but ordering of results is arbitrary.
         '''
-        assert self._state == RUN
+        if self._state != RUN:
+            return
         lost_worker_timeout = lost_worker_timeout or self.lost_worker_timeout
         if chunksize == 1:
             result = IMapUnorderedIterator(self._cache,
@@ -871,7 +1072,8 @@ class Pool(object):
             ...     callback(retval)
 
         '''
-        assert self._state == RUN
+        if self._state != RUN:
+            return
         lost_worker_timeout = lost_worker_timeout or self.lost_worker_timeout
         if soft_timeout and SIG_SOFT_TIMEOUT is None:
             warnings.warn(UserWarning("Soft timeouts are not supported: "
@@ -887,8 +1089,11 @@ class Pool(object):
             if timeout or soft_timeout:
                 # start the timeout handler thread when required.
                 self._start_timeout_handler()
-            self._taskqueue.put(([(result._job, None,
-                                   func, args, kwds)], None))
+            if self.with_task_thread:
+                self._taskqueue.put(([(result._job, None,
+                                       func, args, kwds)], None))
+            else:
+                self._quick_put((result._job, None, func, args, kwds))
             return result
 
     def map_async(self, func, iterable, chunksize=None, callback=None,
@@ -904,7 +1109,8 @@ class Pool(object):
         '''
         Helper function to implement map, starmap and their async counterparts.
         '''
-        assert self._state == RUN
+        if self._state != RUN:
+            return
         if not hasattr(iterable, '__len__'):
             iterable = list(iterable)
 
@@ -940,11 +1146,11 @@ class Pool(object):
         debug('closing pool')
         if self._state == RUN:
             self._state = CLOSE
-            self._worker_handler.close()
-            self._taskqueue.put(None)
-            self._worker_handler.join()
             if self._putlock:
                 self._putlock.clear()
+            self._worker_handler.close()
+            self._taskqueue.put(None)
+            stop_if_not_current(self._worker_handler)
 
     def terminate(self):
         debug('terminating pool')
@@ -955,11 +1161,11 @@ class Pool(object):
     def join(self):
         assert self._state in (CLOSE, TERMINATE)
         debug('joining worker handler')
-        self._worker_handler.join()
+        stop_if_not_current(self._worker_handler)
         debug('joining task handler')
-        self._task_handler.join()
+        self._task_handler.stop()
         debug('joining result handler')
-        self._result_handler.join()
+        self._result_handler.stop()
         debug('result handler joined')
         for i, p in enumerate(self._pool):
             debug('joining worker %s/%s (%r)', i, len(self._pool), p)
@@ -1008,14 +1214,14 @@ class Pool(object):
                     p.terminate()
 
         debug('joining task handler')
-        task_handler.join()
+        task_handler.stop()
 
         debug('joining result handler')
-        result_handler.join()
+        result_handler.stop()
 
         if timeout_handler is not None:
             debug('joining timeout handler')
-            timeout_handler.join(1e100)
+            timeout_handler.stop(1e100)
 
         if pool and hasattr(pool[0], 'terminate'):
             debug('joining pool workers')
